<?php

namespace FiloBlu\Flow\Model\From;

use Exception;
use FiloBlu\Flow\Helper\Data;
use FiloBlu\Flow\Helper\EmailNotification;
use FiloBlu\Flow\Helper\LoggerProvider;
use Magento\Catalog\Model\ProductFactory;
use Magento\CatalogInventory\Api\StockConfigurationInterface;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\CatalogInventory\Model\Stock as CatalogInventoryStock;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ObjectManager;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\DB\Adapter\AdapterInterface;
use Magento\Framework\Exception\MailException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Indexer\ActionFactory;
use Magento\Framework\Indexer\IndexerRegistry;
use Magento\Framework\Model\Context;
use Magento\Framework\Registry;
use Magento\InventoryApi\Api\Data\SourceItemInterface;
use Magento\InventoryApi\Api\SourceItemsSaveInterface;
use Monolog\Logger;
use Zend_Db;

/**
 * @see https://devdocs.magento.com/guides/v2.3/inventory/
 * Class Stock
 * @package FiloBlu\Flow\Model\From
 */
class Stock extends AbstractFrom
{
    /**
     * @var string[]
     */
    const SOURCES = ['snatt', 'default', 'magento', 've'];

    /**
     * @var Logger
     */
    protected $_logger;

    /**
     * @var Data
     */
    protected $helperFlow;

    /**
     * @var ProductFactory
     */
    protected $productFactory;

    /**
     * @var StockRegistryInterface
     */
    protected $stockRegistry;

    /**
     * @var \FiloBlu\Flow\Helper\Stock
     */
    protected $_stock_qty;

    /**
     * @var ProductMetadataInterface
     */
    protected $_productMetadata;

    /**
     * @var EmailNotification
     */
    protected $emailNotification;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var
     */
    protected $connection;

    /**
     * @var StockConfigurationInterface
     */
    protected $stockConfiguration;

    /**
     * @var SourceItemsSaveInterface
     */
    protected $sourceItemSave;
    /**
     * @var IndexerRegistry
     */
    protected $indexerRegistry;
    /**
     * @var ActionFactory
     */
    protected $actionFactory;

    /**
     * Stock constructor.
     * @param Context $context
     * @param Registry $registry
     * @param Data $helper_flow
     * @param LoggerProvider $loggerProvider
     * @param ProductFactory $productFactory
     * @param StockRegistryInterface $stockRegistry
     * @param StockConfigurationInterface $stockConfiguration
     * @param \FiloBlu\Flow\Helper\Stock $stock_qty
     * @param ProductMetadataInterface $productMetadata
     * @param EmailNotification $emailNotification
     * @param ScopeConfigInterface $scopeConfig
     * @param IndexerRegistry $indexerRegistry
     * @param ActionFactory $actionFactory
     * @param array $data
     */
    public function __construct(
        Context $context,
        Registry $registry,
        Data $helper_flow,
        LoggerProvider $loggerProvider,
        ProductFactory $productFactory,
        StockRegistryInterface $stockRegistry,
        StockConfigurationInterface $stockConfiguration,
        \FiloBlu\Flow\Helper\Stock $stock_qty,
        ProductMetadataInterface $productMetadata,
        EmailNotification $emailNotification,
        ScopeConfigInterface $scopeConfig,
        IndexerRegistry $indexerRegistry,
        ActionFactory $actionFactory,
        array $data = []
    ) {
        $this->helperFlow = $helper_flow;
        $this->_logger = $loggerProvider->getLogger();
        $this->productFactory = $productFactory;
        $this->stockRegistry = $stockRegistry;
        $this->_stock_qty = $stock_qty;
        $this->_productMetadata = $productMetadata;
        $this->emailNotification = $emailNotification;
        $this->scopeConfig = $scopeConfig;
        $this->stockConfiguration = $stockConfiguration;
        $this->indexerRegistry = $indexerRegistry;
        $this->actionFactory = $actionFactory;
        parent::__construct($context, $registry, null, null, $data);
    }

    /**
     * @return void
     */
    public function _construct()
    {
        $this->_init(\FiloBlu\Flow\Model\ResourceModel\From\Stock::class);
    }

    /**
     * @return $this|bool|mixed
     * @throws Exception
     */
    public function process()
    {
        if (version_compare($this->_productMetadata->getVersion(), '2.3.0', '>=')) {
            $this->process23();
            return $this;
        }

        $this->process22();
        return $this;
    }

    /**
     *
     */
    protected function process23()
    {
        $processReservedStock = $this->getChannel()->getChannelConfig()->getProcessReservedStock();

        $sku = $this->getData('sku');
        $sourceCode = $this->getData('magazzino');
        $quantity = intval($this->getData('fisico_disponibile'));

        try {
            $sourceItem = $this->createSourceItemInterface();
            $sourceItem->setSourceCode($sourceCode);
            $sourceItem->setSku($sku);

            if ($processReservedStock) {
                $reservedQty = $this->getReservedQuantity23($sku);
                $originalQuantity = $quantity;

                if ($reservedQty !== 0) {
                    if ($reservedQty < 0) {
                        $quantity += abs($reservedQty);
                    } else {
                        $this->_logger->critical("Stock reservation error for product {$sku} - Reserved quantity sum = {$reservedQty}, Original quantity {$originalQuantity}");
                    }
                }
            }

            $sourceItem->setQuantity($quantity);

            if ($quantity > 0) {
                $sourceItem->setStatus(SourceItemInterface::STATUS_IN_STOCK);
            } else {
                $sourceItem->setStatus(SourceItemInterface::STATUS_OUT_OF_STOCK);
            }

            $this->getSourceItemsSave()->execute([$sourceItem]);
            $this->_logger->info("Process STOCK - qty {$quantity} SKU: {$sku}");
        } catch (Exception $e) {
            $this->_logger->info('Errore nel salvataggio dello Stock. ' . $e->getMessage());
        }

        return $this;
    }

    /**
     * Create a SourceItem
     * @return SourceItemInterface
     */
    public function createSourceItemInterface()
    {
        return ObjectManager::getInstance()->create(SourceItemInterface::class);
    }

    /**
     * FOS-720
     *
     * @param $productId integer
     * @return int
     */
    public function getReservedQuantity23($productId)
    {
        $readConnection = $this->getConnection();

        $query = $readConnection->select()
            ->from(['ir' => 'inventory_reservation'], ['SUM(ir.quantity) as qty'])
            ->where("ir.sku = '$productId'");

        $results = $readConnection->fetchOne($query);

        if (!$results) {
            return 0;
        }

        return (int)$results;
    }

    /**
     * @return false|AdapterInterface
     */
    protected function getConnection()
    {
        if ($this->connection) {
            return $this->connection;
        }

        return ($this->connection = $this->getResource()->getConnection());
    }

    /**
     * @return SourceItemsSaveInterface
     */
    public function getSourceItemsSave()
    {
        if ($this->sourceItemSave === null) {
            $this->sourceItemSave = ObjectManager::getInstance()->create(SourceItemsSaveInterface::class);
        }

        return $this->sourceItemSave;
    }

    /**
     * @return bool
     * @throws Exception
     */
    protected function process22()
    {
        $connection = $this->getConnection();

        $sku = $this->getData('sku');
        $productId = $this->productFactory->create()->getIdBySku($sku);

        if (!$productId) {
            return $this->error(false, "Product not found! product_code: '{$sku}'");
        }

        $stock = $this->stockRegistry->getStockItem($productId);

        try {
            $raw_qty = intval($this->getData('fisico_disponibile'));
            $qty = $this->_stock_qty->getRealQty($productId, $raw_qty);

            // Se non viene ritornata una quantità numerica ma una stringa vuol dire che c'è stato un errore
            if (!is_numeric($qty)) {
                // Salvo il messaggio relativo all'errore
                $this->addEmailNotifyError($qty);
                // Dopo aver salvato il messaggio aggiorno la quantità da impostare a 0
                $qty = 0;
            }

            $order_list = $this->_stock_qty->getReservedQtyOrderList($productId);

            $stock->setQty($qty);

            /* Leave the stock inventory status untouched for products having backorder flag set */

            if ($stock->getBackorders() === CatalogInventoryStock::BACKORDERS_NO) {
                if ($qty > $stock->getMinQty()) {
                    $stock->setIsInStock(true);
                } else {
                    $stock->setIsInStock(false);
                }
            }

            $stockData = $stock->getData();

            if (isset($stockData['type_id'])) {
                unset($stockData['type_id']);
            }

            if (isset($stockData['stock_status_changed_automatically_flag'])) {
                unset($stockData['stock_status_changed_automatically_flag']);
            }

            $connection->insertOnDuplicate('cataloginventory_stock_item', $stockData);

            if ($raw_qty !== $qty) {
                $preserved_qty = $raw_qty - $qty;
                $this->error(false, "Preserved Quantity '{$preserved_qty}' For Product ID '{$productId}' SKU '{$sku}' (From orders: '{$order_list}')");
            }

            $this->_logger->info("Process STOCK - qty {$qty} ID: {$productId} SKU: {$sku}");
        } catch (Exception $e) {
            $this->_logger->info('Errore nel salvataggio dello Stock. ' . $e->getMessage());
        }
    }

    /**
     * @param $file
     * @return mixed|void
     * @throws MailException
     * @throws NoSuchEntityException
     */
    public function sendErrorNotifications($file)
    {

        // Se sono stati rilevati errori e salvati i relativi messaggi si procede con la composizione della mail
        if (!empty($this->getEmailNotifyError())) {

            // Recupero gli indirizzi email dei destinatari
            $emailAddressString = $this->scopeConfig->getValue(self::XML_PATH_EMAIL_ADDRESS_NOTIFY_ERROR);

            if ($emailAddressString && $emailAddressString !== '') {
                $emailAddress = explode(',', $emailAddressString);
                $emailAddress = array_map('trim', $emailAddress);

                $storeName = $this->scopeConfig->getValue(self::XML_PATH_STORE_NAME);
                // Compongo il testo del messaggio con i vari errori
                $errorString = '<p>' . implode('</p><p>', $this->getEmailNotifyError()) . '</p>';

                // Array con i dati che serviranno a comporre la mail
                $data = [
                    'mail_subject' => 'Reporting breakdowns in stock loading for ' . $storeName,
                    'file_name' => $file->getName(),
                    'errors' => $errorString
                ];

                // Invio email di notifica
                $this->emailNotification->notify($emailAddress, $data);
            }
        }
    }

    /**
     * @param $query
     * @return bool
     */
    protected function postprocessFileRows($query)
    {
        // TODO : check also against module MSI. We need to check if MSI in sot uninstalled or disabled
        if (version_compare($this->_productMetadata->getVersion(), '2.3.0', '>=')) {
            return $this->postprocessFileRows23();
        }

        return $this->postprocessFileRows22();
    }

    /**
     * Workaround until stock reindex will be fixed
     */
    protected function postprocessFileRows23()
    {
        $indexer = $this->indexerRegistry->get('cataloginventory_stock');
        $action = $this->actionFactory->create($indexer->getActionClass());
        $action->executeFull();
    }

    /**
     * @return bool
     */
    protected function postprocessFileRows22()
    {
        try {
            $this->updateConfigurablesStockStatus();
        } catch (Exception $e) {
            $this->_logger->info('Errore nel salvataggio del Post Stock Item Fix Tuning. ' . $e->getMessage());
            return false;
        }

        return true;
    }

    /**
     * 1] $this->stockConfiguration->getBackorders($this->getStoreId()) = 0
     *      backorders = 0 && use_config_backorders = 1
     *
     * 2]  backorders = 0 && use_config_backorders = 0
     */
    protected function updateConfigurablesStockStatus()
    {
        $connection = $this->getConnection();
        $entityField = $this->getEntityField();

        $sql = $connection->select()
            ->from(['csi' => 'cataloginventory_stock_item'], [])
            ->joinLeft(['cpe' => 'catalog_product_entity'], 'cpe.entity_id = csi.product_id', [])
            ->joinLeft(['cpsl' => 'catalog_product_super_link'], 'cpe.entity_id = cpsl.product_id', ['cpsl.parent_id'])
            ->where('cpsl.parent_id is not null')
            ->group('cpsl.parent_id');

        $outOfStockSql = clone $sql;
        $inStockSql = clone $sql;

        $outOfStockSql->having('SUM(csi.qty) <= 0');
        $inStockSql->having('SUM(csi.qty) > 0');

        /* 1] Update Out Of Stock configurables */
        $products = $connection->select()
            ->from('catalog_product_entity', ['entity_id'])
            ->where('type_id = ?', ['configurable'])
            ->where("{$entityField} IN (?)", [$outOfStockSql])
            ->query()->fetchAll(Zend_Db::FETCH_NUM);

        $configurableIds = [];

        foreach ($products as $row) {
            $configurableIds[] = $row[0];
        }

        $ids = implode(',', $configurableIds);
        $connection->query("UPDATE cataloginventory_stock_item SET is_in_stock = 0 WHERE `product_id` IN ({$ids})");

        /* 2] Update In Stock configurables */
        $products = $products = $connection->select()
            ->from('catalog_product_entity', ['entity_id'])
            ->where('type_id = ?', ['configurable'])
            ->where("{$entityField} IN (?)", [$inStockSql])
            ->query()->fetchAll(Zend_Db::FETCH_NUM);

        $configurableIds = [];

        foreach ($products as $row) {
            $configurableIds[] = $row[0];
        }

        $ids = implode(',', $configurableIds);
        $connection->query("UPDATE cataloginventory_stock_item SET is_in_stock = 1 WHERE `product_id` IN ({$ids})");
    }

    /**
     * @return string
     */
    protected function getEntityField()
    {
        $edition = $this->_productMetadata->getEdition();
        return ($edition === 'Enterprise' || $edition === 'B2B') ? 'row_id' : 'entity_id';
    }
}
